@@ -5,13 +5,17 @@
 
 use rand::{thread_rng, Rng};
 use secrecy::Zeroize;
+use std::env;
 use std::fmt;
 use std::io::{self, BufRead, BufReader, Read, Write};
 use std::iter;
 use std::path::Path;
 use std::process::{ChildStdin, ChildStdout, Command, Stdio};
 
-use crate::format::{grease_the_joint, read, write, Stanza};
+use crate::{
+    format::{grease_the_joint, read, write, Stanza},
+    io::{DebugReader, DebugWriter},
+};
 
 pub const IDENTITY_V1: &str = "identity-v1";
 pub const RECIPIENT_V1: &str = "recipient-v1";
@@ -25,12 +29,14 @@ const RESPONSE_UNSUPPORTED: &str = "unsupported";
 #[derive(Debug)]
 pub enum Error {
     Fail,
+    Unsupported,
 }
 
 impl fmt::Display for Error {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         match self {
             Self::Fail => write!(f, "General plugin protocol error"),
+            Self::Unsupported => write!(f, "Unsupported command"),
         }
     }
 }
@@ -59,19 +65,31 @@ pub struct Connection<R: Read, W: Write> {
     _working_dir: Option<tempfile::TempDir>,
 }
 
-impl Connection<ChildStdout, ChildStdin> {
-    /// Start a plugin binary with the given state machine.
+impl Connection<DebugReader<ChildStdout>, DebugWriter<ChildStdin>> {
+    /// Starts a plugin binary with the given state machine.
+    ///
+    /// If the `AGEDEBUG` environment variable is set to `plugin`, then all messages sent
+    /// to and from the plugin, as well as anything the plugin prints to its `stderr`,
+    /// will be printed to the `stderr` of the parent process.
     pub fn open(binary: &Path, state_machine: &str) -> io::Result<Self> {
         let working_dir = tempfile::tempdir()?;
+        let debug_enabled = env::var("AGEDEBUG").map(|s| s == "plugin").unwrap_or(false);
         let process = Command::new(binary.canonicalize()?)
             .arg(format!("--age-plugin={}", state_machine))
             .current_dir(working_dir.path())
             .stdin(Stdio::piped())
             .stdout(Stdio::piped())
-            .stderr(Stdio::piped())
+            .stderr(if debug_enabled {
+                Stdio::inherit()
+            } else {
+                Stdio::null()
+            })
             .spawn()?;
-        let input = BufReader::new(process.stdout.expect("could open stdout"));
-        let output = process.stdin.expect("could open stdin");
+        let input = BufReader::new(DebugReader::new(
+            process.stdout.expect("could open stdout"),
+            debug_enabled,
+        ));
+        let output = DebugWriter::new(process.stdin.expect("could open stdin"), debug_enabled);
         Ok(Connection {
             input,
             output,
@@ -319,6 +337,7 @@ impl<'a, R: Read, W: Write> BidirSend<'a, R, W> {
         match s.tag.as_ref() {
             RESPONSE_OK => Ok(Ok(s)),
             RESPONSE_FAIL => Ok(Err(Error::Fail)),
+            RESPONSE_UNSUPPORTED => Ok(Err(Error::Unsupported)),
             tag => Err(io::Error::new(
                 io::ErrorKind::InvalidData,
                 format!("unexpected response: {}", tag),
@@ -342,6 +361,7 @@ impl<'a, R: Read, W: Write> BidirSend<'a, R, W> {
         match s.tag.as_ref() {
             RESPONSE_OK => Ok(Ok(s)),
             RESPONSE_FAIL => Ok(Err(Error::Fail)),
+            RESPONSE_UNSUPPORTED => Ok(Err(Error::Unsupported)),
             tag => Err(io::Error::new(
                 io::ErrorKind::InvalidData,
                 format!("unexpected response: {}", tag),
@@ -362,6 +382,11 @@ impl<'a, R: Read, W: Write> Reply<'a, R, W> {
         )
     }
 
+    /// Reply with `ok`, metadata, and optional data.
+    pub fn ok_with_metadata<S: AsRef<str>>(self, metadata: &[S], data: Option<&[u8]>) -> Response {
+        Response(self.0.send(RESPONSE_OK, metadata, data.unwrap_or_default()))
+    }
+
     /// The command failed (for example, the user failed to respond to an input request).
     pub fn fail(self) -> Response {
         Response(self.0.send::<&str>(RESPONSE_FAIL, &[], &[]))
